/*
  Copyright 2008-2020 David Robillard <d@drobilla.net>
  Copyright 1999-2000 Paul Kellett (Maxim Digital Audio)
  SPDX-License-Identifier: GPL-2.0-or-later or MIT

  This is free software: you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  This software is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this software. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __mdaTestTone_H
#define __mdaTestTone_H

#include "audioeffectx.h"

#include <cstdint>

class mdaTestTone : public AudioEffectX
{
public:
	mdaTestTone(audioMasterCallback audioMaster);
	~mdaTestTone();

	virtual void process(float **inputs, float **outputs, int32_t sampleFrames);
	virtual void processReplacing(float **inputs, float **outputs, int32_t sampleFrames);
	virtual void setProgramName(char *name);
	virtual void getProgramName(char *name);
	virtual bool getProgramNameIndexed (int32_t category, int32_t index, char* name);
	virtual void setParameter(int32_t index, float value);
	virtual float getParameter(int32_t index);
	virtual void getParameterLabel(int32_t index, char *label);
	virtual void getParameterDisplay(int32_t index, char *text);
	virtual void getParameterName(int32_t index, char *text);
  virtual void iso2string(float b, char *text);
  virtual void suspend();
  void midi2string(float n, char *text);

	virtual bool getEffectName(char *name);
	virtual bool getVendorString(char *text);
	virtual bool getProductString(char *text);
	virtual int32_t getVendorVersion() { return 1000; }

protected:
  int updateTx, updateRx;
  void update();

	float fParam0, fParam1, fParam2;
	float fParam3, fParam4, fParam5;
	float fParam6, fParam7;

  float thru, left, right, len;
  float zz0, zz1, zz2, zz3, zz4, zz5, phi, dphi;
  float sw, swd, swx, fscale, cal, calx;
  int32_t swt;
  int mode;

	char programName[32], disp1[16], disp2[16];
};

#endif
